"""
Command-line utility to help process GitHub Actions.
"""

import argparse
import json
import os
from typing import Any, Callable, Dict, List, Optional


def generate_argument_parser() -> argparse.ArgumentParser:
    commands = ["type", "initial", "terminal", "repo"]
    parser = argparse.ArgumentParser(description="Locust GitHub Actions helper")
    parser.add_argument(
        "command",
        choices=commands,
        help="Information you would like this helper to provide",
    )
    return parser


def helper_push(command: str, event: Dict[str, Any]) -> str:
    """
    Event structure defined here:
    https://docs.github.com/en/free-pro-team@latest/developers/webhooks-and-events/webhook-events-and-payloads#push
    """
    if command == "initial":
        return event["before"]
    elif command == "terminal":
        return event["after"]
    elif command == "repo":
        return event["repository"]["html_url"]

    raise Exception(f"Unknown command: {command}")


def helper_pr(command: str, event: Dict[str, Any]) -> str:
    """
    Event structure defined here:
    https://docs.github.com/en/free-pro-team@latest/developers/webhooks-and-events/webhook-events-and-payloads#pull_request
    """
    pull_request = event["pull_request"]
    if command == "initial":
        return pull_request["base"]["sha"]
    elif command == "terminal":
        return pull_request["head"]["sha"]
    elif command == "repo":
        return pull_request["head"]["repo"]["html_url"]

    raise Exception(f"Unknown command: {command}")


def main():
    parser = generate_argument_parser()
    args = parser.parse_args()

    helpers: Dict[str, Callable[[str, Dict[str, Any]], str]] = {
        "push": helper_push,
        "pull_request": helper_pr,
        "pull_request_target": helper_pr,
    }
    github_event_name: str = os.environ.get("GITHUB_EVENT_NAME", "")
    if github_event_name not in helpers:
        raise ValueError(
            f"Helper only works with events of the following types: {','.join(helpers)}"
        )

    if args.command == "type":
        print(github_event_name)
        return

    github_event_path: Optional[str] = os.environ.get("GITHUB_EVENT_PATH", None)
    if github_event_path is None:
        raise ValueError(
            "No path was provided for the GitHub event (GITHUB_EVENT_PATH not set)"
        )

    with open(github_event_path, "r") as ifp:
        event: Dict[str, Any] = json.load(ifp)

    print(helpers[github_event_name](args.command, event))
    return


if __name__ == "__main__":
    main()



"""
git-related functionality
"""
import argparse
import json
import os
import sys
from typing import Any, List, Optional, Tuple

from pydantic import BaseModel
import pygit2


class GitRepositoryNotFound(Exception):
    """
    Raised when a git repository was not found where one was expected.
    """


class LineInfo(BaseModel):
    old_line_number: int
    new_line_number: int
    line_type: str
    line: str


class HunkInfo(BaseModel):
    header: str
    lines: List[LineInfo]


class PatchInfo(BaseModel):
    old_file: str
    new_file: str
    hunks: List[HunkInfo]


class RunResponse(BaseModel):
    repo: str
    initial_ref: str
    terminal_ref: Optional[str]
    patches: List[PatchInfo]


def get_repository(path: str = ".") -> pygit2.Repository:
    """
    Returns a git repository object if it can find one at the given path, otherwise raises a
    GitRepositoryNotFound error.
    """
    repository_path: Optional[str] = pygit2.discover_repository(path)
    if repository_path is None:
        raise GitRepositoryNotFound(f"No git repository found at path: {path}")
    return pygit2.Repository(repository_path)


def get_patches(
    repository: pygit2.Repository,
    initial: Optional[str] = None,
    terminal: Optional[str] = None,
) -> List[PatchInfo]:
    """
    Returns a list of patches taking the given repository from the initial revision to the terminal
    one.
    """
    diff = repository.diff(a=initial, b=terminal, context_lines=0)
    return [
        PatchInfo(
            old_file=patch.delta.old_file.path,
            new_file=patch.delta.new_file.path,
            hunks=[process_hunk(hunk) for hunk in patch.hunks],
        )
        for patch in diff
    ]


def process_hunk(hunk: pygit2.DiffHunk) -> HunkInfo:
    """
    Processes a hunk from a git diff into a HunkInfo object.
    """
    return HunkInfo(
        header=hunk.header,
        lines=[
            LineInfo(
                old_line_number=line.old_lineno,
                new_line_number=line.new_lineno,
                line_type=line.origin,
                line=line.content,
            )
            for line in hunk.lines
        ],
    )


def revision_file(
    repository: pygit2.Repository, revision: Optional[str], filepath: str
) -> bytes:
    """
    Returns the bytes from the file at the given filepath on the given revision. If revision is
    None, returns the bytes from the filepath in the current working tree.

    Filepath is expected to be an absolute, normalized path.
    """
    repo_path = os.path.normpath(repository.workdir)

    content = bytes()

    if revision is None:
        assert (
            os.path.commonpath([repo_path, filepath]) == repo_path
        ), f"File ({filepath}) is not contained in repository ({repo_path})"
        with open(filepath, "rb") as ifp:
            content = ifp.read()
    else:
        relative_path = os.path.relpath(filepath, repo_path)
        dirname, basename = os.path.split(relative_path)
        components: List[str] = [basename]
        while dirname:
            dirname, basename = os.path.split(dirname)
            components.append(basename)
        components.reverse()

        commit = repository.revparse_single(revision)
        current_tree = commit.tree
        for component in components:
            current_tree = current_tree[component]
        content = current_tree.data

    return content


def populate_argument_parser(parser: argparse.ArgumentParser) -> None:
    """
    Populates an argparse ArgumentParser object with the commonly used arguments for this module.

    Mutates the provided parser.
    """
    parser.add_argument(
        "-r", "--repo", required=False, default=".", help="Path to git repository"
    )
    parser.add_argument(
        "initial",
        nargs="?",
        default=None,
        help="Initial git revision",
    )
    parser.add_argument(
        "terminal",
        nargs="?",
        default=None,
        help="Terminal git revision",
    )


def run(repo_dir: str, initial: Optional[str], terminal: Optional[str]) -> RunResponse:
    repo = get_repository(repo_dir)

    initial_ref = repo.revparse_single("HEAD").short_id
    if initial is not None:
        initial_ref = repo.revparse_single(initial).short_id

    terminal_ref = None
    if terminal is not None:
        terminal_ref = repo.revparse_single(terminal).short_id

    patches = get_patches(repo, initial_ref, terminal_ref)
    response = RunResponse(
        repo=os.path.normpath(repo.workdir),
        initial_ref=initial_ref,
        terminal_ref=terminal_ref,
        patches=patches,
    )
    return response


def main():
    parser = argparse.ArgumentParser(description="Locust: git functionality")
    populate_argument_parser(parser)
    parser.add_argument(
        "-o",
        "--output",
        required=False,
        type=argparse.FileType("w"),
        default=sys.stdout,
        help=(
            "Path to which to write list of PatchInfo objects generated by this module (in JSON "
            "format)"
        ),
    )

    args = parser.parse_args()

    response = run(args.repo, args.initial, args.terminal)

    try:
        with args.output as ofp:
            print(response.json(), file=ofp)
    except BrokenPipeError:
        pass


if __name__ == "__main__":
    main()



"""
The Locust CLI
"""
import argparse
import json
import os
import sys
from typing import Any, Dict, Optional

from . import git
from . import parse
from . import render
from . import version


def generate_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Locust: Analyze Python code across git references",
        epilog=f"Version {version.LOCUST_VERSION}",
    )
    git.populate_argument_parser(parser)
    render.populate_argument_parser(parser)
    parser.add_argument(
        "-o",
        "--output",
        required=False,
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="Path to which to write results",
    )
    return parser


def main():
    parser = generate_argument_parser()
    args = parser.parse_args()

    git_result = git.run(args.repo, args.initial, args.terminal)

    parse_result = parse.run(git_result)

    results_string = render.run(parse_result, args.format, args.github)

    try:
        with args.output as ofp:
            print(results_string, file=ofp)
    except BrokenPipeError:
        pass


if __name__ == "__main__":
    main()



"""
AST-related functionality
"""
import argparse
import ast
import json
import os
import sys
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel
from pygit2 import Repository

from . import git


def hunk_boundary(
    hunk: git.HunkInfo, operation_type: Optional[str] = None
) -> Optional[Tuple[int, int]]:
    """
    Calculates boundary for the given hunk, returning a tuple of the form:
    (<line number of boundary start>, <line number of boundary end>)

    If operation_type is provided, it is used to filter down only to lines whose line_type matches
    the operation_type. Possible values: "+", "-", None.

    If there are no lines of the  given type in the hunk, returns None.
    """
    line_type_p = lambda line: True
    if operation_type is not None:
        line_type_p = lambda line: line.line_type == operation_type

    admissible_lines = [line for line in hunk.lines if line_type_p(line)]

    if not admissible_lines:
        return None

    return (admissible_lines[0].new_line_number, admissible_lines[-1].new_line_number)


class RawDefinition(BaseModel):
    name: str
    change_type: str
    line: int
    offset: int
    end_line: Optional[int] = None
    end_offset: Optional[int] = None
    parent: Optional[Tuple[str, int]] = None


class LocustChange(BaseModel):
    name: str
    change_type: str
    filepath: str
    revision: Optional[str]
    line: int
    changed_lines: int
    total_lines: Optional[int] = None
    parent: Optional[Tuple[str, int]] = None


class LocustVisitor(ast.NodeVisitor):
    FUNCTION_DEF = "function"
    ASYNC_FUNCTION_DEF = "async_function"
    CLASS_DEF = "class"

    def __init__(
        self,
        repository: Repository,
        revision: Optional[str],
        patches: List[git.PatchInfo],
    ):
        self.repository = repository
        self.revision = None
        if revision is not None:
            self.revision = repository.revparse_single(revision).short_id

        self.insertion_boundaries: Dict[str, List[Tuple[int, int]]] = {}
        for patch in patches:
            _, extension = os.path.splitext(patch.new_file)
            if extension != ".py":
                continue
            patch_filepath = os.path.realpath(
                os.path.join(repository.workdir, patch.new_file)
            )
            raw_insertion_boundaries = [
                hunk_boundary(hunk, "+") for hunk in patch.hunks
            ]
            self.insertion_boundaries[os.path.abspath(patch_filepath)] = [
                boundary
                for boundary in raw_insertion_boundaries
                if boundary is not None
            ]

        self.scope: List[Tuple[str, int, Optional[int]]] = []
        self.definitions: List[RawDefinition] = []
        self.imports: Dict[str, str] = {}

    def _visit_class_or_function_def(
        self,
        node: Union[ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef],
        def_type: str,
    ) -> None:
        self.scope = [
            spec for spec in self.scope if spec[2] is not None and spec[2] > node.lineno
        ]
        self.scope.append((node.name, node.lineno, node.end_lineno))
        parent: Optional[Tuple[str, int]] = None
        if len(self.scope) > 1:
            parent = (
                ".".join([spec[0] for spec in self.scope[:-1]]),
                self.scope[-2][1],
            )
        self.definitions.append(
            RawDefinition(
                name=".".join([spec[0] for spec in self.scope]),
                change_type=def_type,
                line=node.lineno,
                offset=node.col_offset,
                end_line=node.end_lineno,
                end_offset=node.end_col_offset,
                parent=parent,
            )
        )
        self.generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self._visit_class_or_function_def(node, self.FUNCTION_DEF)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self._visit_class_or_function_def(node, self.ASYNC_FUNCTION_DEF)

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self._visit_class_or_function_def(node, self.CLASS_DEF)

    def reset(self):
        self.scope = []
        self.definitions = []
        self.imports = {}

    def parse(self, filepath: str) -> List[LocustChange]:
        abs_filepath = os.path.realpath(os.path.abspath(filepath))
        if (
            abs_filepath not in self.insertion_boundaries
            or not self.insertion_boundaries[abs_filepath]
        ):
            return []

        insertion_boundaries = sorted(
            self.insertion_boundaries[abs_filepath], key=lambda p: p[0]
        )

        source_bytes = git.revision_file(self.repository, self.revision, filepath)
        source = source_bytes.decode()

        root = ast.parse(source)
        self.reset()
        self.visit(root)

        locust_changes: List[LocustChange] = []
        for definition in self.definitions:
            possible_boundaries = [
                boundary
                for boundary in insertion_boundaries
                if definition.end_line is None or boundary[0] <= definition.end_line
            ]
            if not possible_boundaries:
                continue

            total_lines: Optional[int] = None
            if definition.end_line is not None:
                total_lines = definition.end_line - definition.line + 1

            candidate_insertion = max(
                possible_boundaries,
                key=lambda p: p[0],
            )

            if candidate_insertion[1] >= definition.line:
                changed_lines = 0
                for start, end in possible_boundaries:
                    if (end >= definition.line) and (
                        definition.end_line is None or start <= definition.end_line
                    ):
                        end_line = end
                        if (
                            definition.end_line is not None
                            and definition.end_line < end
                        ):
                            end_line = definition.end_line

                        changed_lines += end_line - max(start, definition.line) + 1

                locust_changes.append(
                    LocustChange(
                        name=definition.name,
                        change_type=definition.change_type,
                        filepath=filepath,
                        revision=self.revision,
                        line=definition.line,
                        changed_lines=changed_lines,
                        total_lines=total_lines,
                        parent=definition.parent,
                    )
                )

        return locust_changes

    def parse_all(self) -> List[LocustChange]:
        changed_changes: List[LocustChange] = []
        for filepath in self.insertion_boundaries:
            changed_changes.extend(self.parse(filepath))
        return changed_changes


class RunResponse(BaseModel):
    repo: str
    initial_ref: str
    terminal_ref: Optional[str]
    patches: List[git.PatchInfo]
    changes: List[LocustChange]


def run(git_result: git.RunResponse) -> RunResponse:
    repo = git.get_repository(git_result.repo)
    visitor = LocustVisitor(repo, git_result.terminal_ref, git_result.patches)
    changes = visitor.parse_all()
    return RunResponse(
        repo=git_result.repo,
        initial_ref=git_result.initial_ref,
        terminal_ref=git_result.terminal_ref,
        patches=git_result.patches,
        changes=changes,
    )


def main():
    parser = argparse.ArgumentParser(description="Locust: Python parsing functionality")
    parser.add_argument(
        "-i",
        "--input",
        type=argparse.FileType("r"),
        default=sys.stdin,
        help="Path to git result. If not specified, reads from stdin.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="Path to write parse results to (in JSON format)",
    )

    args = parser.parse_args()

    with args.input as ifp:
        git_result_json = json.load(ifp)
        git_result = git.RunResponse.parse_obj(git_result_json)

    result = run(git_result)

    try:
        with args.output as ofp:
            print(result.json(), file=ofp)
    except BrokenPipeError:
        pass


if __name__ == "__main__":
    main()
